from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Iterator,
    List,
    Optional,
    Type,
    TypeVar,
    AsyncIterator,
)
from pydantic import BaseModel
import polars as pl

T = TypeVar("T", bound=BaseModel)


class PaginatedIterator(Generic[T], Iterator[T]):
    def __init__(
        self,
        client: Any,
        endpoint: str,
        params: Optional[Dict[str, Any]] = None,
        *,
        item_class: Optional[Type[T]] = None,
        item_parser: Optional[Callable[[Dict[str, Any]], T]] = None,
        df_transform: Optional[Callable[[T], List[Dict[str, Any]]]] = None,
    ):
        self.client = client
        self.endpoint = endpoint
        self.params = params.copy() if params else {}
        self.buffer: list[T] = []
        self.cursor: Optional[str] = None
        self.finished: bool = False
        self._df_transform = df_transform

        # Provide a default parser using Pydantic if not given.
        if item_parser is None:
            if item_class is None:
                raise ValueError(
                    "Either 'item_class' or 'item_parser' must be provided"
                )
            self.item_parser = lambda item: item_class.model_validate(item)
        else:
            self.item_parser = item_parser

    def _fetch_next_page(self) -> None:
        if self.cursor:
            self.params["cursor"] = self.cursor
        response = self.client._request_sync("GET", self.endpoint, params=self.params)
        raw_results = response.get("results", [])
        self.buffer = [self.item_parser(item) for item in raw_results]
        pagination = response.get("pagination", {})
        self.cursor = pagination.get("cursor")
        # Mark finished if it's the last page or no items are returned.
        self.finished = pagination.get("last_page", True) or not self.buffer

    def __iter__(self) -> "PaginatedIterator[T]":
        return self

    def __next__(self) -> T:
        if not self.buffer and not self.finished:
            self._fetch_next_page()
        if self.buffer:
            return self.buffer.pop(0)
        raise StopIteration

    def fetch_page(self) -> list[T]:
        """
        Fetches the next page of results and returns them as a list.
        """
        if not self.buffer and not self.finished:
            self._fetch_next_page()
        page = self.buffer.copy()
        self.buffer.clear()  # Clear the buffer after returning the page.
        return page

    def to_polars(self) -> pl.DataFrame:
        """
        Fetches all items from all pages, applies an optional transformation,
        and returns them as a Polars DataFrame.
        This method will consume the iterator.
        """
        all_items_collected: list[T] = []

        # Consume items currently in the buffer from any previous partial iteration
        all_items_collected.extend(self.buffer)
        self.buffer.clear()

        # Fetch and collect all remaining items from subsequent pages
        while not self.finished:
            self._fetch_next_page()  # Fetches new page into self.buffer, updates self.finished
            all_items_collected.extend(self.buffer)
            self.buffer.clear()  # Clear buffer after processing its contents

        if not all_items_collected:
            return pl.DataFrame()  # Return empty DataFrame if no data

        processed_data_for_df: list[dict[str, Any]] = []
        if self._df_transform:
            for item_model in all_items_collected:
                transformed_result = self._df_transform(item_model)
                processed_data_for_df.extend(transformed_result)
        else:
            # Default behavior: convert each Pydantic model to a dict
            for item_model in all_items_collected:
                processed_data_for_df.append(item_model.model_dump())

        return pl.from_dicts(processed_data_for_df)


class AsyncPaginatedIterator(Generic[T], AsyncIterator[T]):
    def __init__(
        self,
        client: Any,
        endpoint: str,
        params: Optional[Dict[str, Any]] = None,
        *,
        item_class: Optional[Type[T]] = None,
        item_parser: Optional[Callable[[Dict[str, Any]], T]] = None,
        df_transform: Optional[Callable[[T], List[Dict[str, Any]]]] = None,
    ):
        self.client = client
        self.endpoint = endpoint
        self.params = params.copy() if params else {}
        self.buffer: list[T] = []
        self.cursor: Optional[str] = None
        self.finished: bool = False
        self._df_transform = df_transform

        if item_parser is None:
            if item_class is None:
                raise ValueError(
                    "Either 'item_class' or 'item_parser' must be provided"
                )
            self.item_parser = lambda item: item_class.model_validate(item)
        else:
            self.item_parser = item_parser

    async def _fetch_next_page(self) -> None:
        if self.cursor:
            self.params["cursor"] = self.cursor
        response = await self.client._request_async(
            "GET", self.endpoint, params=self.params
        )
        raw_results = response.get("results", [])
        self.buffer = [self.item_parser(item) for item in raw_results]
        pagination = response.get("pagination", {})
        self.cursor = pagination.get("cursor")
        self.finished = pagination.get("last_page", True) or not self.buffer

    def __aiter__(self) -> "AsyncPaginatedIterator[T]":
        return self

    async def __anext__(self) -> T:
        if not self.buffer and not self.finished:
            await self._fetch_next_page()
        if self.buffer:
            return self.buffer.pop(0)
        raise StopAsyncIteration

    async def afetch_page(self) -> list[T]:
        """
        Asynchronously fetches the next page of results and returns them as a list.
        """
        if not self.buffer and not self.finished:
            await self._fetch_next_page()
        page = self.buffer.copy()
        self.buffer.clear()
        return page

    async def to_polars(self) -> pl.DataFrame:
        """
        Asynchronously fetches all items from all pages, applies an optional transformation,
        and returns them as a Polars DataFrame.
        This method will consume the iterator.
        """
        all_items_collected: list[T] = []

        # Consume items currently in the buffer from any previous partial iteration
        all_items_collected.extend(self.buffer)
        self.buffer.clear()

        # Fetch and collect all remaining items from subsequent pages
        while not self.finished:
            await (
                self._fetch_next_page()
            )  # Fetches new page into self.buffer, updates self.finished
            all_items_collected.extend(self.buffer)
            self.buffer.clear()  # Clear buffer after processing its contents

        if not all_items_collected:
            return pl.DataFrame()  # Return empty DataFrame if no data

        processed_data_for_df: list[dict[str, Any]] = []
        if self._df_transform:
            for item_model in all_items_collected:
                transformed_result = self._df_transform(item_model)
                processed_data_for_df.extend(transformed_result)
        else:
            # Default behavior: convert each Pydantic model to a dict
            for item_model in all_items_collected:
                processed_data_for_df.append(item_model.model_dump())

        return pl.from_dicts(processed_data_for_df)



from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Type,
    TypeVar,
    Iterator,
    AsyncIterator,
    Self,
)
from pydantic import BaseModel
import polars as pl

T = TypeVar("T", bound=BaseModel)


class StaticListIterator(Generic[T], Iterator[T], AsyncIterator[T]):
    """
    Helper to fetch, parse, and iterate over a list of items from an API endpoint
    that returns the full list under a 'results' key without pagination.
    """

    def __init__(
        self,
        client: Any,
        endpoint: str,
        params: Optional[Dict[str, Any]] = None,
        *,
        item_class: Optional[Type[T]] = None,
        item_parser: Optional[Callable[[Dict[str, Any]], T]] = None,
    ):
        self.client = client
        self.endpoint = endpoint
        self.params = params.copy() if params else {}

        if item_parser is None:
            if item_class is None:
                raise ValueError(
                    "Either 'item_class' or 'item_parser' must be provided"
                )
            self.item_parser: Callable[[Dict[str, Any]], T] = (
                lambda item: item_class.model_validate(item)
            )
        else:
            self.item_parser = item_parser

        # Internal state for iteration
        self._fetched_data: Optional[List[T]] = None
        self._iter_index: int = 0

    def _fetch_sync_if_needed(self) -> None:
        """Fetches data synchronously if not already fetched."""
        if self._fetched_data is None:
            self._fetched_data = self.fetch_all()  # Use the existing fetch method
            self._iter_index = 0  # Reset index when fetching

    def fetch_all(self) -> List[T]:
        """
        Synchronously fetches all results from the endpoint and parses them.
        If data has already been fetched for iteration, returns the cached data.

        Returns:
            A list of parsed items of type T.
        """
        if self._fetched_data is None:
            # Fetch and store internally
            response = self.client._request_sync(
                "GET", self.endpoint, params=self.params
            )
            raw_results = response.get("results", [])
            if not isinstance(raw_results, list):
                raise TypeError(
                    f"Expected 'results' field to be a list, got {type(raw_results)}"
                )
            self._fetched_data = [self.item_parser(item) for item in raw_results]
            self._iter_index = 0  # Ensure index is reset

        # Return a copy to prevent external modification of internal state
        return self._fetched_data.copy()

    def __iter__(self) -> Self:
        """Returns the iterator object itself, fetching data if needed."""
        self._fetch_sync_if_needed()
        return self

    def __next__(self) -> T:
        """Returns the next item in the fetched list."""
        self._fetch_sync_if_needed()  # Ensure data is fetched
        if self._fetched_data is not None and self._iter_index < len(
            self._fetched_data
        ):
            item = self._fetched_data[self._iter_index]
            self._iter_index += 1
            return item
        else:
            raise StopIteration

    async def _afetch_async_if_needed(self) -> None:
        """Fetches data asynchronously if not already fetched."""
        if self._fetched_data is None:
            self._fetched_data = (
                await self.afetch_all()
            )  # Use the existing async fetch method
            self._iter_index = 0  # Reset index

    async def afetch_all(self) -> List[T]:
        """
        Asynchronously fetches all results from the endpoint and parses them.
        If data has already been fetched for iteration, returns the cached data.

        Returns:
            A list of parsed items of type T.
        """
        if self._fetched_data is None:
            # Fetch and store internally
            response = await self.client._request_async(
                "GET", self.endpoint, params=self.params
            )
            raw_results = response.get("results", [])
            if not isinstance(raw_results, list):
                raise TypeError(
                    f"Expected 'results' field to be a list, got {type(raw_results)}"
                )
            self._fetched_data = [self.item_parser(item) for item in raw_results]
            self._iter_index = 0  # Ensure index is reset

        # Return a copy
        return self._fetched_data.copy()

    def __aiter__(self) -> Self:
        """Returns the async iterator object itself."""
        return self

    async def __anext__(self) -> T:
        """Returns the next item, fetching data asynchronously if needed."""
        await self._afetch_async_if_needed()  # Ensure data is fetched
        if self._fetched_data is not None and self._iter_index < len(
            self._fetched_data
        ):
            item = self._fetched_data[self._iter_index]
            self._iter_index += 1
            return item
        else:
            raise StopAsyncIteration

    def to_polars(self) -> pl.DataFrame:
        """
        Fetches all results (if not already fetched) and converts them
        into a Polars DataFrame.

        Assumes that the generic type T is a Pydantic model.

        Returns:
            A Polars DataFrame containing the fetched data.
        """
        self._fetch_sync_if_needed()  # Ensure data is available

        if self._fetched_data is None or not self._fetched_data:
            return pl.DataFrame()  # Return empty DataFrame if no data

        # Convert list of Pydantic models to list of dictionaries
        data_dicts = [item.model_dump() for item in self._fetched_data]

        # Create Polars DataFrame from the list of dictionaries
        df = pl.from_dicts(data_dicts)
        return df



from datetime import datetime


class APIError(Exception):
    """
    General exception for API errors.
    Attributes:
        message: The error message.
        code: HTTP status code.
        status: HTTP status text.
        timestamp: When the error occurred.
    """

    def __init__(self, message, code=None, status=None, timestamp=None):
        self.message = message
        self.code = code
        self.status = status or "Error"
        self.timestamp = timestamp or datetime.now()
        super().__init__(f"{self.status}: {self.message}")


class RateLimitError(APIError):
    """
    Exception raised when rate limiting is exceeded.
    """

    def __init__(self, status=None, timestamp=None):
        self.message = "Rate limit exceeded"
        self.code = 429
        self.status = status or "Error"
        self.timestamp = timestamp or datetime.now()
        super().__init__(f"{self.status}: {self.message}")



from .types import Pathway
from .base import BaseClient


class Climate:
    def __init__(self, client: BaseClient):
        self.client = client

    def list_horizons(self) -> list[int]:
        """
        List the available horizons for climate analysis.
        """
        return list(range(2025, 2100, 5))

    def list_pathways(self) -> list[Pathway]:
        """
        List the available pathways for climate analysis.
        """
        return [
            "SV",
            "ssp126",
            "ssp245",
            "ssp370",
            "ssp585",
            "<2 degrees",
            "2-3 degrees",
            "3-4 degrees",
            ">4 degrees",
        ]



from typing import Any, Dict, Optional
from .base import BaseClient
from .types import (
    AssetClimateScore,
    AssetImpactScore,
    AssetTypeClimateScore,
    AssetTypeImpactScore,
    CountryClimateScore,
    CountryImpactScore,
    ImpactScore,
    MarketIndex,
    Company,
    ClimateScore,
    Pathway,
    HorizonYear,
)
from .pagination import PaginatedIterator, AsyncPaginatedIterator
from .static_list import StaticListIterator


class Markets:
    def __init__(self, client: BaseClient):
        self.client = client

    def search_indexes(
        self,
        *,
        name: Optional[str] = None,
        **extra_params: Any,
    ) -> list[MarketIndex]:
        """
        Search for market indexes by name.

        Parameters:
            name (Optional[str]): The name of the market index to search for.
            **extra_params (Any): Additional parameters to pass to the API.

        Returns:
            list[MarketIndex]: A list of MarketIndex objects matching the search criteria.
        """
        params: Dict[str, Any] = {}
        if name is not None:
            params["name"] = name
        params.update(extra_params)
        response = self.client._request_sync(
            "GET", "/markets/indexes/search", params=params
        )
        results = [MarketIndex.model_validate(item) for item in response["results"]]
        return results

    async def search_indexes_async(
        self,
        *,
        name: Optional[str] = None,
        **extra_params: Any,
    ) -> list[MarketIndex]:
        """
        Search for market indexes by name asynchronously.

        Parameters:
            name (Optional[str]): The name of the market index to search for.
            **extra_params (Any): Additional parameters to pass to the API.

        Returns:
            list[MarketIndex]: A list of MarketIndex objects matching the search criteria.
        """
        params: Dict[str, Any] = {}
        if name is not None:
            params["name"] = name
        params.update(extra_params)
        response = await self.client._request_async(
            "GET", "/markets/indexes/search", params=params
        )
        results = [MarketIndex.model_validate(item) for item in response["results"]]
        return results

    def list_indexes(self) -> PaginatedIterator[MarketIndex]:
        """
        List all market indexes.

        Returns:
            PaginatedIterator[MarketIndex]: An iterator over MarketIndex objects.
        """
        return PaginatedIterator(
            self.client, "/markets/indexes", {}, item_class=MarketIndex
        )

    async def list_indexes_async(self) -> AsyncPaginatedIterator[MarketIndex]:
        """
        List all market indexes asynchronously.

        Returns:
            AsyncPaginatedIterator[MarketIndex]: An asynchronous iterator over MarketIndex objects.
        """
        return AsyncPaginatedIterator(
            self.client, "/markets/indexes", {}, item_class=MarketIndex
        )

    def get_index(self, index_id: str) -> MarketIndex:
        """
        Get a market index by its unique ID.

        Parameters:
            index_id (str): The unique identifier of the market index.

        Returns:
            MarketIndex: The MarketIndex object.
        """
        response = self.client._request_sync("GET", f"/markets/indexes/{index_id}")
        return MarketIndex(**response)

    async def get_index_async(self, index_id: str) -> MarketIndex:
        """
        Get a market index by its unique ID asynchronously.

        Parameters:
            index_id (str): The unique identifier of the market index.

        Returns:
            MarketIndex: The MarketIndex object.
        """
        response = await self.client._request_async(
            "GET", f"/markets/indexes/{index_id}"
        )
        return MarketIndex(**response)

    def get_index_companies(self, index_id: str) -> PaginatedIterator[Company]:
        """
        Get all companies in a market index.

        Parameters:
            index_id (str): The unique identifier of the market index.

        Returns:
            PaginatedIterator[Company]: An iterator over Company objects in the index.
        """
        return PaginatedIterator(
            self.client,
            f"/markets/indexes/{index_id}/companies",
            {},
            item_class=Company,
        )

    async def get_index_companies_async(
        self, index_id: str
    ) -> AsyncPaginatedIterator[Company]:
        """
        Get all companies in a market index asynchronously.

        Parameters:
            index_id (str): The unique identifier of the market index.

        Returns:
            AsyncPaginatedIterator[Company]: An asynchronous iterator over Company objects in the index.
        """
        return AsyncPaginatedIterator(
            self.client,
            f"/markets/indexes/{index_id}/companies",
            {},
            item_class=Company,
        )

    def get_index_climate_scores(
        self, index_id: str, pathway: Pathway, horizon: HorizonYear
    ) -> ClimateScore:
        """
        Get the climate scores for a market index.

        Parameters:
            index_id (str): The unique identifier of the market index.
            pathway (Pathway): Climate scenario pathway powered by Climate Earth Digital Twin.
            horizon (HorizonYear): Climatology year representing a decadal period.

        Returns:
            ClimateScore: The ClimateScore object for the market index.
        """
        response = self.client._request_sync(
            "GET",
            f"/markets/indexes/{index_id}/climate/scores",
            params={
                "pathway": pathway,
                "horizon": horizon,
            },
        )
        return ClimateScore(**response)

    async def get_index_climate_scores_async(
        self, index_id: str, pathway: Pathway, horizon: HorizonYear
    ) -> ClimateScore:
        """
        Get the climate scores for a market index asynchronously.

        Parameters:
            index_id (str): The unique identifier of the market index.
            pathway (Pathway): Climate scenario pathway powered by Climate Earth Digital Twin.
            horizon (HorizonYear): Climatology year representing a decadal period.

        Returns:
            ClimateScore: The ClimateScore object for the market index.
        """
        response = await self.client._request_async(
            "GET",
            f"/markets/indexes/{index_id}/climate/scores",
            params={
                "pathway": pathway,
                "horizon": horizon,
            },
        )
        return ClimateScore(**response)

    def get_index_impact_scores(
        self, index_id: str, pathway: Pathway, horizon: HorizonYear
    ) -> StaticListIterator[ImpactScore]:
        """
        Get the impact scores for a market index.

        Parameters:
            index_id (str): The unique identifier of the market index.
            pathway (Pathway): Climate scenario pathway powered by Climate Earth Digital Twin.
            horizon (HorizonYear): Climatology year representing a decadal period.

        Returns:
            StaticListIterator[ImpactScore]: An iterator over ImpactScore objects for the market index.
        """
        return StaticListIterator(
            self.client,
            f"/markets/indexes/{index_id}/climate/impacts",
            {
                "pathway": pathway,
                "horizon": horizon,
            },
            item_class=ImpactScore,
        )

    async def get_index_impact_scores_async(
        self, index_id: str, pathway: Pathway, horizon: HorizonYear
    ) -> StaticListIterator[ImpactScore]:
        """
        Get the impact scores for a market index asynchronously.

        Parameters:
            index_id (str): The unique identifier of the market index.
            pathway (Pathway): Climate scenario pathway powered by Climate Earth Digital Twin.
            horizon (HorizonYear): Climatology year representing a decadal period.

        Returns:
            StaticListIterator[ImpactScore]: An asynchronous iterator over ImpactScore objects for the market index.
        """
        return StaticListIterator(
            self.client,
            f"/markets/indexes/{index_id}/climate/impacts",
            {
                "pathway": pathway,
                "horizon": horizon,
            },
            item_class=ImpactScore,
        )

    def list_index_asset_impact_scores(
        self,
        index_id: str,
        pathway: Pathway,
        horizon: HorizonYear,
        **extra_params: Any,
    ) -> PaginatedIterator[AssetImpactScore]:
        """
        Get the impact scores for all assets of a market index.

        Parameters:
            index_id (str): The unique identifier of the market index.
            pathway (Pathway): Climate scenario pathway powered by Climate Earth Digital Twin.
            horizon (HorizonYear): Climatology year representing a decadal period.
            **extra_params (Any): Additional parameters to pass to the API.

        Returns:
            PaginatedIterator[AssetImpactScore]: An iterator over AssetImpactScore objects for the index's assets.
        """
        params: Dict[str, Any] = {}
        params["pathway"] = pathway
        params["horizon"] = horizon
        params["metric"] = (
            "dcr_score,cvar_99,var_99,cvar_95,var_95,cvar_50,var_50,expected_impact"
        )
        params.update(extra_params)
        return PaginatedIterator(
            self.client,
            f"/markets/indexes/{index_id}/assets/climate/impacts",
            params,
            item_class=AssetImpactScore,
            df_transform=lambda asset: [
                {
                    "asset_id": asset.asset_id,
                    **{
                        f"index_{risk.index_name}": risk.index_impact_cvar_50
                        for risk in asset.index_risks
                    },
                }
            ],
        )

    def list_index_asset_climate_scores(
        self,
        index_id: str,
        pathway: Pathway,
        horizon: HorizonYear,
        **extra_params: Any,
    ) -> PaginatedIterator[AssetClimateScore]:
        """
        Get the climate scores for all assets of a market index.

        Parameters:
            index_id (str): The unique identifier of the market index.
            pathway (Pathway): Climate scenario pathway powered by Climate Earth Digital Twin.
            horizon (HorizonYear): Climatology year representing a decadal period.
            **extra_params (Any): Additional parameters to pass to the API.

        Returns:
            PaginatedIterator[AssetClimateScore]: An iterator over AssetClimateScore objects for the index's assets.
        """
        params: Dict[str, Any] = {}
        params["pathway"] = pathway
        params["horizon"] = horizon
        params["metric"] = (
            "dcr_score,cvar_99,var_99,cvar_95,var_95,cvar_50,var_50,expected_impact"
        )
        params.update(extra_params)
        return PaginatedIterator(
            self.client,
            f"/markets/indexes/{index_id}/assets/climate/scores",
            params,
            item_class=AssetClimateScore,
        )

    async def list_index_asset_climate_scores_async(
        self,
        index_id: str,
        pathway: Pathway,
        horizon: HorizonYear,
        **extra_params: Any,
    ) -> AsyncPaginatedIterator[AssetClimateScore]:
        """
        Get the climate scores for all assets of a market index asynchronously.

        Parameters:
            index_id (str): The unique identifier of the market index.
            pathway (Pathway): Climate scenario pathway powered by Climate Earth Digital Twin.
            horizon (HorizonYear): Climatology year representing a decadal period.
            **extra_params (Any): Additional parameters to pass to the API.

        Returns:
            AsyncPaginatedIterator[AssetClimateScore]: An asynchronous iterator over AssetClimateScore objects for the index's assets.
        """
        params: Dict[str, Any] = {}
        params["pathway"] = pathway
        params["horizon"] = horizon
        params["metric"] = (
            "dcr_score,cvar_99,var_99,cvar_95,var_95,cvar_50,var_50,expected_impact"
        )
        params.update(extra_params)
        return AsyncPaginatedIterator(
            self.client,
            f"/markets/indexes/{index_id}/assets/climate/scores",
            params,
            item_class=AssetClimateScore,
        )

    async def list_index_asset_impact_scores_async(
        self,
        index_id: str,
        pathway: Pathway,
        horizon: HorizonYear,
        **extra_params: Any,
    ) -> AsyncPaginatedIterator[AssetImpactScore]:
        """
        Get the impact scores for all assets of a market index asynchronously.

        Parameters:
            index_id (str): The unique identifier of the market index.
            pathway (Pathway): Climate scenario pathway powered by Climate Earth Digital Twin.
            horizon (HorizonYear): Climatology year representing a decadal period.
            **extra_params (Any): Additional parameters to pass to the API.

        Returns:
            AsyncPaginatedIterator[AssetImpactScore]: An asynchronous iterator over AssetImpactScore objects for the index's assets.
        """
        params: Dict[str, Any] = {}
        params["pathway"] = pathway
        params["horizon"] = horizon
        params["metric"] = (
            "dcr_score,cvar_99,var_99,cvar_95,var_95,cvar_50,var_50,expected_impact"
        )
        params.update(extra_params)
        return AsyncPaginatedIterator(
            self.client,
            f"/markets/indexes/{index_id}/assets/climate/impacts",
            params,
            item_class=AssetImpactScore,
            df_transform=lambda asset: [
                {
                    "asset_id": asset.asset_id,
                    **{
                        f"index_{risk.index_name}": risk.index_impact_cvar_50
                        for risk in asset.index_risks
                    },
                }
            ],
        )

    def aggregate_index_asset_climate_scores_by_country(
        self, index_id: str, pathway: Pathway, horizon: HorizonYear
    ) -> StaticListIterator[CountryClimateScore]:
        """
        Get the climate scores for all assets in a market index aggregated by country.

        Parameters:
            index_id (str): The unique identifier of the market index.
            pathway (Pathway): Climate scenario pathway powered by Climate Earth Digital Twin.
            horizon (HorizonYear): Climatology year representing a decadal period.

        Returns:
            StaticListIterator[CountryClimateScore]: An iterator over CountryClimateScore objects, aggregated by country.
        """
        return StaticListIterator(
            self.client,
            f"/markets/indexes/{index_id}/assets/climate/scores/aggregation",
            {
                "by": "country",
                "pathway": pathway,
                "horizon": horizon,
                "metric": "dcr_score,cvar_99,var_99,cvar_95,var_95,cvar_50,var_50,expected_impact",
            },
            item_class=CountryClimateScore,
        )

    async def aggregate_index_asset_climate_scores_by_country_async(
        self, index_id: str, pathway: Pathway, horizon: HorizonYear
    ) -> StaticListIterator[CountryClimateScore]:
        """
        Get the climate scores for all assets in a market index aggregated by country asynchronously.

        Parameters:
            index_id (str): The unique identifier of the market index.
            pathway (Pathway): Climate scenario pathway powered by Climate Earth Digital Twin.
            horizon (HorizonYear): Climatology year representing a decadal period.

        Returns:
            StaticListIterator[CountryClimateScore]: An asynchronous iterator over CountryClimateScore objects, aggregated by country.
        """
        return StaticListIterator(
            self.client,
            f"/markets/indexes/{index_id}/assets/climate/scores/aggregation",
            {
                "by": "country",
                "pathway": pathway,
                "horizon": horizon,
                "metric": "dcr_score,cvar_99,var_99,cvar_95,var_95,cvar_50,var_50,expected_impact",
            },
            item_class=CountryClimateScore,
        )

    def aggregate_index_asset_impact_scores_by_country(
        self, index_id: str, pathway: Pathway, horizon: HorizonYear
    ) -> StaticListIterator[CountryImpactScore]:
        """
        Get the impact scores for all assets in a market index aggregated by country.

        Parameters:
            index_id (str): The unique identifier of the market index.
            pathway (Pathway): Climate scenario pathway powered by Climate Earth Digital Twin.
            horizon (HorizonYear): Climatology year representing a decadal period.

        Returns:
            StaticListIterator[CountryImpactScore]: An iterator over CountryImpactScore objects, aggregated by country.
        """
        return StaticListIterator(
            self.client,
            f"/markets/indexes/{index_id}/assets/climate/impacts/aggregation",
            {
                "by": "country",
                "pathway": pathway,
                "horizon": horizon,
                "metric": "dcr_score,cvar_99,var_99,cvar_95,var_95,cvar_50,var_50,expected_impact",
            },
            item_class=CountryImpactScore,
        )

    async def aggregate_index_asset_impact_scores_by_country_async(
        self, index_id: str, pathway: Pathway, horizon: HorizonYear
    ) -> StaticListIterator[CountryImpactScore]:
        """
        Get the impact scores for all assets in a market index aggregated by country asynchronously.

        Parameters:
            index_id (str): The unique identifier of the market index.
            pathway (Pathway): Climate scenario pathway powered by Climate Earth Digital Twin.
            horizon (HorizonYear): Climatology year representing a decadal period.

        Returns:
            StaticListIterator[CountryImpactScore]: An asynchronous iterator over CountryImpactScore objects, aggregated by country.
        """
        return StaticListIterator(
            self.client,
            f"/markets/indexes/{index_id}/assets/climate/impacts/aggregation",
            {
                "by": "country",
                "pathway": pathway,
                "horizon": horizon,
                "metric": "dcr_score,cvar_99,var_99,cvar_95,var_95,cvar_50,var_50,expected_impact",
            },
            item_class=CountryImpactScore,
        )

    def aggregate_index_asset_climate_scores_by_asset_type(
        self, index_id: str, pathway: Pathway, horizon: HorizonYear
    ) -> StaticListIterator[AssetTypeClimateScore]:
        """
        Get the climate scores for all assets in a market index aggregated by asset type.

        Parameters:
            index_id (str): The unique identifier of the market index.
            pathway (Pathway): Climate scenario pathway powered by Climate Earth Digital Twin.
            horizon (HorizonYear): Climatology year representing a decadal period.

        Returns:
            StaticListIterator[AssetTypeClimateScore]: An iterator over AssetTypeClimateScore objects, aggregated by asset type.
        """
        return StaticListIterator(
            self.client,
            f"/markets/indexes/{index_id}/assets/climate/scores/aggregation",
            {
                "by": "asset_type",
                "pathway": pathway,
                "horizon": horizon,
                "metric": "dcr_score,cvar_99,var_99,cvar_95,var_95,cvar_50,var_50,expected_impact",
            },
            item_class=AssetTypeClimateScore,
        )

    async def aggregate_index_asset_climate_scores_by_asset_type_async(
        self, index_id: str, pathway: Pathway, horizon: HorizonYear
    ) -> StaticListIterator[AssetTypeClimateScore]:
        """
        Get the climate scores for all assets in a market index aggregated by asset type asynchronously.

        Parameters:
            index_id (str): The unique identifier of the market index.
            pathway (Pathway): Climate scenario pathway powered by Climate Earth Digital Twin.
            horizon (HorizonYear): Climatology year representing a decadal period.

        Returns:
            StaticListIterator[AssetTypeClimateScore]: An asynchronous iterator over AssetTypeClimateScore objects, aggregated by asset type.
        """
        return StaticListIterator(
            self.client,
            f"/markets/indexes/{index_id}/assets/climate/scores/aggregation",
            {
                "by": "asset_type",
                "pathway": pathway,
                "horizon": horizon,
                "metric": "dcr_score,cvar_99,var_99,cvar_95,var_95,cvar_50,var_50,expected_impact",
            },
            item_class=AssetTypeClimateScore,
        )

    def aggregate_index_asset_impact_scores_by_asset_type(
        self, index_id: str, pathway: Pathway, horizon: HorizonYear
    ) -> StaticListIterator[AssetTypeImpactScore]:
        """
        Get the impact scores for all assets in a market index aggregated by asset type.

        Parameters:
            index_id (str): The unique identifier of the market index.
            pathway (Pathway): Climate scenario pathway powered by Climate Earth Digital Twin.
            horizon (HorizonYear): Climatology year representing a decadal period.

        Returns:
            StaticListIterator[AssetTypeImpactScore]: An iterator over AssetTypeImpactScore objects, aggregated by asset type.
        """
        return StaticListIterator(
            self.client,
            f"/markets/indexes/{index_id}/assets/climate/impacts/aggregation",
            {
                "by": "asset_type",
                "pathway": pathway,
                "horizon": horizon,
                "metric": "dcr_score,cvar_99,var_99,cvar_95,var_95,cvar_50,var_50,expected_impact",
            },
            item_class=AssetTypeImpactScore,
        )

    async def aggregate_index_asset_impact_scores_by_asset_type_async(
        self, index_id: str, pathway: Pathway, horizon: HorizonYear
    ) -> StaticListIterator[AssetTypeImpactScore]:
        """
        Get the impact scores for all assets in a market index aggregated by asset type asynchronously.

        Parameters:
            index_id (str): The unique identifier of the market index.
            pathway (Pathway): Climate scenario pathway powered by Climate Earth Digital Twin.
            horizon (HorizonYear): Climatology year representing a decadal period.

        Returns:
            StaticListIterator[AssetTypeImpactScore]: An asynchronous iterator over AssetTypeImpactScore objects, aggregated by asset type.
        """
        return StaticListIterator(
            self.client,
            f"/markets/indexes/{index_id}/assets/climate/impacts/aggregation",
            {
                "by": "asset_type",
                "pathway": pathway,
                "horizon": horizon,
                "metric": "dcr_score,cvar_99,var_99,cvar_95,var_95,cvar_50,var_50,expected_impact",
            },
            item_class=AssetTypeImpactScore,
        )



from typing import Any, Dict
import backoff
import httpx
import os

from .errors import RateLimitError, APIError

BASE_URL = "https://api.riskthinking.ai/v3"


class BaseClient:
    def __init__(
        self,
        api_key: str | None = None,
        timeout: float = 15.0,
        base_url: str | None = None,
    ):
        if api_key is None:
            api_key = os.getenv("RISKTHINKING_API_KEY", None)
        if api_key is None:
            raise Exception("API key is required to initialize the sdk")

        self._api_key = api_key

        self._base_url = BASE_URL if not base_url else base_url
        if not self._base_url.endswith("/v3"):
            self._base_url = str(self._base_url) + "/v3"

        headers = {
            "Authorization": f"Bearer {self._api_key}",
            "Content-Type": "application/json",
        }
        self._sync_client = httpx.Client(
            base_url=self._base_url, headers=headers, timeout=timeout
        )
        self._async_client = httpx.AsyncClient(
            base_url=self._base_url, headers=headers, timeout=timeout
        )

    @backoff.on_exception(backoff.expo, RateLimitError, max_tries=5)
    def _request_sync(self, method: str, path: str, **kwargs) -> Dict[str, Any]:
        # Remove the /v3 prefix since it's part of the base URL
        if path.startswith("/v3"):
            path = path[3:]
        response = self._sync_client.request(method, path, **kwargs)
        return self._handle_response_sync(response)

    @backoff.on_exception(backoff.expo, RateLimitError, max_tries=5)
    async def _request_async(self, method: str, path: str, **kwargs) -> Dict[str, Any]:
        # Remove the /v3 prefix since it's part of the base URL
        if path.startswith("/v3"):
            path = path[3:]
        response = await self._async_client.request(method, path, **kwargs)
        return await self._handle_response_async(response)

    def _handle_response_sync(self, response: httpx.Response) -> Dict[str, Any]:
        if response.status_code == 429:
            raise RateLimitError("Rate limit exceeded")
        if response.status_code >= 400:
            try:
                error = response.json().get("error", {})
            except Exception:
                error = {}
            raise APIError(
                message=error.get("message", "Unknown error"),
                code=error.get("code", response.status_code),
                status=error.get("status", f"{response.status_code} Error"),
                timestamp=error.get("timestamp"),
            )
        return response.json()

    async def _handle_response_async(self, response: httpx.Response) -> Dict[str, Any]:
        if response.status_code == 429:
            raise RateLimitError("Rate limit exceeded")
        if response.status_code >= 400:
            try:
                json_data = await response.json()
                error = json_data.get("error", {})
            except Exception:
                error = {}
            raise APIError(
                message=error.get("message", "Unknown error"),
                code=error.get("code", response.status_code),
                status=error.get("status", f"{response.status_code} Error"),
                timestamp=error.get("timestamp"),
            )
        return await response.json()

    # Generic HTTP methods
    def get(
        self, path: str, params: Dict[str, Any] = dict(), **kwargs
    ) -> Dict[str, Any]:
        """
        Make a GET request to the API.
        The version /v3 is automatically added to the path.

        Args:
            path: API endpoint path
            params: Query parameters
            **kwargs: Additional arguments to pass to the request

        Returns:
            API response data as JSON deserialized into a dictionary
        """
        return self._request_sync("GET", path, params=params, **kwargs)

    async def get_async(
        self, path: str, params: Dict[str, Any] = dict(), **kwargs
    ) -> Dict[str, Any]:
        """
        Make an asynchronous GET request to the API.
        The version /v3 is automatically added to the path.

        Args:
            path: API endpoint path
            params: Query parameters
            **kwargs: Additional arguments to pass to the request

        Returns:
            API response data as JSON deserialized into a dictionary
        """
        return await self._request_async("GET", path, params=params, **kwargs)

    def post(
        self, path: str, json: Dict[str, Any] = dict(), **kwargs
    ) -> Dict[str, Any]:
        """
        Make a POST request to the API.
        The version /v3 is automatically added to the path.
        Args:
            path: API endpoint path
            json: JSON body data
            **kwargs: Additional arguments to pass to the request

        Returns:
            API response data as JSON deserialized into a dictionary
        """
        return self._request_sync("POST", path, json=json, **kwargs)

    async def post_async(
        self, path: str, json: Dict[str, Any] = dict(), **kwargs
    ) -> Dict[str, Any]:
        """
        Make an asynchronous POST request to the API.
        The version /v3 is automatically added to the path.

        Args:
            path: API endpoint path
            json: JSON body data
            **kwargs: Additional arguments to pass to the request

        Returns:
            API response data as JSON deserialized into a dictionary
        """
        return await self._request_async("POST", path, json=json, **kwargs)

    def put(self, path: str, json: Dict[str, Any] = dict(), **kwargs) -> Dict[str, Any]:
        """
        Make a PUT request to the API.
        The version /v3 is automatically added to the path.

        Args:
            path: API endpoint path
            json: JSON body data
            **kwargs: Additional arguments to pass to the request

        Returns:
            API response data as JSON deserialized into a dictionary
        """
        return self._request_sync("PUT", path, json=json, **kwargs)

    async def put_async(
        self, path: str, json: Dict[str, Any] = dict(), **kwargs
    ) -> Dict[str, Any]:
        """
        Make an asynchronous PUT request to the API.
        The version /v3 is automatically added to the path.

        Args:
            path: API endpoint path
            json: JSON body data
            **kwargs: Additional arguments to pass to the request

        Returns:
            API response data as JSON deserialized into a dictionary
        """
        return await self._request_async("PUT", path, json=json, **kwargs)

    def delete(self, path: str, **kwargs) -> Dict[str, Any]:
        """
        Make a DELETE request to the API.
        The version /v3 is automatically added to the path.

        Args:
            path: API endpoint path
            **kwargs: Additional arguments to pass to the request

        Returns:
            API response data as JSON deserialized into a dictionary
        """
        return self._request_sync("DELETE", path, **kwargs)

    async def delete_async(self, path: str, **kwargs) -> Dict[str, Any]:
        """
        Make an asynchronous DELETE request to the API.
        The version /v3 is automatically added to the path.

        Args:
            path: API endpoint path
            **kwargs: Additional arguments to pass to the request

        Returns:
            API response data as JSON deserialized into a dictionary
        """
        return await self._request_async("DELETE", path, **kwargs)

    def patch(
        self, path: str, json: Dict[str, Any] = dict(), **kwargs
    ) -> Dict[str, Any]:
        """
        Make a PATCH request to the API.
        The version /v3 is automatically added to the path.

        Args:
            path: API endpoint path
            json: JSON body data
            **kwargs: Additional arguments to pass to the request

        Returns:
            API response data as JSON deserialized into a dictionary
        """
        return self._request_sync("PATCH", path, json=json, **kwargs)

    async def patch_async(
        self, path: str, json: Dict[str, Any] = dict(), **kwargs
    ) -> Dict[str, Any]:
        """
        Make an asynchronous PATCH request to the API.
        The version /v3 is automatically added to the path.

        Args:
            path: API endpoint path
            json: JSON body data
            **kwargs: Additional arguments to pass to the request

        Returns:
            API response data as JSON deserialized into a dictionary
        """
        return await self._request_async("PATCH", path, json=json, **kwargs)



from typing import Any, Dict, Literal, Optional

from .base import BaseClient
from .types import (
    AssetClimateScore,
    AssetTypeClimateScore,
    AssetTypeImpactScore,
    Company,
    Asset,
    ClimateScore,
    ImpactScore,
    CountryClimateScore,
    CountryImpactScore,
    AssetImpactScore,
    Pathway,
    HorizonYear,
)
from .pagination import PaginatedIterator, AsyncPaginatedIterator
from .static_list import StaticListIterator


class Companies:
    def __init__(self, client: BaseClient):
        self.client = client

    def get_company(self, company_id: str) -> Company:
        """
        Get a company by its unique ID.

        Parameters:
            company_id (str): The unique identifier of the company.

        Returns:
            Company: The Company object.
        """
        response = self.client._request_sync("GET", f"/companies/{company_id}")
        return Company(**response)

    async def get_company_async(self, company_id: str) -> Company:
        """
        Get a company by its unique ID asynchronously.

        Parameters:
            company_id (str): The unique identifier of the company.

        Returns:
            Company: The Company object.
        """
        response = await self.client._request_async("GET", f"/companies/{company_id}")
        return Company(**response)

    def list_companies(
        self,
        *,
        scope: Literal["public", "organization"] = "public",
        **extra_params: Any,
    ) -> PaginatedIterator[Company]:
        """
        List all companies.

        Parameters:
            scope (Literal["public", "organization"]): The scope to filter companies by
                   "public" is the default scope and searches all available companies in VELO.
                   "organization" searches all private companies uploaded to the organization.
            **extra_params (Any): Additional parameters to pass to the API.

        Returns:
            PaginatedIterator[Company]: An iterator over Company objects.
        """
        params: Dict[str, Any] = {}
        params["scope"] = scope
        params.update(extra_params)
        return PaginatedIterator(self.client, "/companies", params, item_class=Company)

    async def list_companies_async(
        self,
        *,
        scope: Literal["public", "organization"] = "public",
        **extra_params: Any,
    ) -> AsyncPaginatedIterator[Company]:
        """
        List all companies asynchronously.

        Parameters:
            scope (Literal["public", "organization"]): The scope to filter companies by
                   "public" is the default scope and searches all available companies in VELO.
                   "organization" searches all private companies uploaded to the organization.
            **extra_params (Any): Additional parameters to pass to the API.

        Returns:
            AsyncPaginatedIterator[Company]: An asynchronous iterator over Company objects.
        """
        params: Dict[str, Any] = {}
        params["scope"] = scope
        params.update(extra_params)
        return AsyncPaginatedIterator(
            self.client, "/companies", params, item_class=Company
        )

    def search_companies(
        self,
        *,
        name: Optional[str] = None,
        **extra_params: Any,
    ) -> list[Company]:
        """
        Search for companies by name.

        Parameters:
            name (Optional[str]): The name of the company to search for.
            **extra_params (Any): Additional parameters to pass to the API.

        Returns:
            list[Company]: A list of Company objects matching the search criteria.
        """
        params: Dict[str, Any] = {}
        if name is not None:
            params["name"] = name
        params.update(extra_params)
        response = self.client._request_sync("GET", "/companies/search", params=params)
        results = [Company.model_validate(item) for item in response["results"]]
        return results

    async def search_companies_async(
        self,
        *,
        name: Optional[str] = None,
        **extra_params: Any,
    ) -> list[Company]:
        """
        Search for companies by name asynchronously.

        Parameters:
            name (Optional[str]): The name of the company to search for.
            **extra_params (Any): Additional parameters to pass to the API.

        Returns:
            list[Company]: A list of Company objects matching the search criteria.
        """
        params: Dict[str, Any] = {}
        if name is not None:
            params["name"] = name
        params.update(extra_params)
        response = await self.client._request_async(
            "GET", "/companies/search", params=params
        )
        results = [Company.model_validate(item) for item in response["results"]]
        return results

    def list_company_assets(
        self,
        company_id: str,
        **extra_params: Any,
    ) -> PaginatedIterator[Asset]:
        """
        List all assets for a company.

        Parameters:
            company_id (str): The unique identifier of the company.
            **extra_params (Any): Additional parameters to pass to the API.

        Returns:
            PaginatedIterator[Asset]: An iterator over Asset objects belonging to the company.
        """
        return PaginatedIterator(
            self.client,
            f"/companies/{company_id}/assets",
            extra_params,
            item_class=Asset,
        )

    async def list_company_assets_async(
        self,
        company_id: str,
        **extra_params: Any,
    ) -> AsyncPaginatedIterator[Asset]:
        """
        List all assets for a company asynchronously.

        Parameters:
            company_id (str): The unique identifier of the company.
            **extra_params (Any): Additional parameters to pass to the API.

        Returns:
            AsyncPaginatedIterator[Asset]: An asynchronous iterator over Asset objects belonging to the company.
        """
        return AsyncPaginatedIterator(
            self.client,
            f"/companies/{company_id}/assets",
            extra_params,
            item_class=Asset,
        )

    def list_uninsurable_company_assets(
        self,
        company_id: str,
        pathway: Pathway,
        horizon: HorizonYear,
        **extra_params: Any,
    ) -> PaginatedIterator[AssetClimateScore]:
        """
        List all uninsurable assets for a company.
        Uninsurable assets are defined as those with cvar_95 >= 0.35.

        Parameters:
            company_id (str): The unique identifier of the company.
            pathway (Pathway): Climate scenario pathway powered by Climate Earth Digital Twin.
            horizon (HorizonYear): Climatology year representing a decadal period.
            **extra_params (Any): Additional parameters to pass to the API.

        Returns:
            PaginatedIterator[AssetClimateScore]: An iterator over AssetClimateScore objects for uninsurable assets.
        """
        params: Dict[str, Any] = {}
        params["pathway"] = pathway
        params["horizon"] = horizon
        params["metric"] = "cvar_95"
        params["min_risk"] = 0.35
        params.update(extra_params)
        return PaginatedIterator(
            self.client,
            f"/companies/{company_id}/assets/climate/scores",
            params,
            item_class=AssetClimateScore,
        )

    async def list_uninsurable_company_assets_async(
        self,
        company_id: str,
        pathway: Pathway,
        horizon: HorizonYear,
        **extra_params: Any,
    ) -> AsyncPaginatedIterator[AssetClimateScore]:
        """
        List all uninsurable assets for a company asynchronously.
        Uninsurable assets are defined as those with cvar_95 >= 0.35.

        Parameters:
            company_id (str): The unique identifier of the company.
            pathway (Pathway): Climate scenario pathway powered by Climate Earth Digital Twin.
            horizon (HorizonYear): Climatology year representing a decadal period.
            **extra_params (Any): Additional parameters to pass to the API.

        Returns:
            AsyncPaginatedIterator[AssetClimateScore]: An asynchronous iterator over AssetClimateScore objects for uninsurable assets.
        """
        params: Dict[str, Any] = {}
        params["pathway"] = pathway
        params["horizon"] = horizon
        params["metric"] = "cvar_95"
        params["min_risk"] = 0.35
        params.update(extra_params)
        return AsyncPaginatedIterator(
            self.client,
            f"/companies/{company_id}/assets/climate/scores",
            params,
            item_class=AssetClimateScore,
        )

    def list_stranded_company_assets(
        self,
        company_id: str,
        pathway: Pathway,
        horizon: HorizonYear,
        **extra_params: Any,
    ) -> PaginatedIterator[AssetClimateScore]:
        """
        List all stranded assets for a company.
        Stranded assets are defined as those with cvar_95 >= 0.75.

        Parameters:
            company_id (str): The unique identifier of the company.
            pathway (Pathway): Climate scenario pathway powered by Climate Earth Digital Twin.
            horizon (HorizonYear): Climatology year representing a decadal period.
            **extra_params (Any): Additional parameters to pass to the API.

        Returns:
            PaginatedIterator[AssetClimateScore]: An iterator over AssetClimateScore objects for stranded assets.
        """
        params: Dict[str, Any] = {}
        params["pathway"] = pathway
        params["horizon"] = horizon
        params["metric"] = "cvar_95"
        params["min_risk"] = 0.75
        params.update(extra_params)
        return PaginatedIterator(
            self.client,
            f"/companies/{company_id}/assets/climate/scores",
            params,
            item_class=AssetClimateScore,
        )

    async def list_stranded_company_assets_async(
        self,
        company_id: str,
        pathway: Pathway,
        horizon: HorizonYear,
        **extra_params: Any,
    ) -> AsyncPaginatedIterator[AssetClimateScore]:
        """
        List all stranded assets for a company asynchronously.
        Stranded assets are defined as those with cvar_95 >= 0.75.

        Parameters:
            company_id (str): The unique identifier of the company.
            pathway (Pathway): Climate scenario pathway powered by Climate Earth Digital Twin.
            horizon (HorizonYear): Climatology year representing a decadal period.
            **extra_params (Any): Additional parameters to pass to the API.

        Returns:
            AsyncPaginatedIterator[AssetClimateScore]: An asynchronous iterator over AssetClimateScore objects for stranded assets.
        """
        params: Dict[str, Any] = {}
        params["pathway"] = pathway
        params["horizon"] = horizon
        params["metric"] = "cvar_95"
        params["min_risk"] = 0.75
        params.update(extra_params)
        return AsyncPaginatedIterator(
            self.client,
            f"/companies/{company_id}/assets/climate/scores",
            params,
            item_class=AssetClimateScore,
        )

    def get_company_climate_scores(
        self, company_id: str, pathway: Pathway, horizon: HorizonYear
    ) -> ClimateScore:
        """
        Get the climate scores for a company.

        Parameters:
            company_id (str): The unique identifier of the company.
            pathway (Pathway): Climate scenario pathway powered by Climate Earth Digital Twin.
            horizon (HorizonYear): Climatology year representing a decadal period.

        Returns:
            ClimateScore: The ClimateScore object for the company.
        """
        response = self.client._request_sync(
            "GET",
            f"/companies/{company_id}/climate/scores",
            params={
                "pathway": pathway,
                "horizon": horizon,
            },
        )
        return ClimateScore(**response)

    async def get_company_climate_scores_async(
        self, company_id: str, pathway: Pathway, horizon: HorizonYear
    ) -> ClimateScore:
        """
        Get the climate scores for a company asynchronously.

        Parameters:
            company_id (str): The unique identifier of the company.
            pathway (Pathway): Climate scenario pathway powered by Climate Earth Digital Twin.
            horizon (HorizonYear): Climatology year representing a decadal period.

        Returns:
            ClimateScore: The ClimateScore object for the company.
        """
        response = await self.client._request_async(
            "GET",
            f"/companies/{company_id}/climate/scores",
            params={
                "pathway": pathway,
                "horizon": horizon,
            },
        )
        return ClimateScore(**response)

    def get_company_impact_scores(
        self, company_id: str, pathway: Pathway, horizon: HorizonYear
    ) -> StaticListIterator[ImpactScore]:
        """
        Get the impact scores for a company.

        Parameters:
            company_id (str): The unique identifier of the company.
            pathway (Pathway): Climate scenario pathway powered by Climate Earth Digital Twin.
            horizon (HorizonYear): Climatology year representing a decadal period.

        Returns:
            StaticListIterator[ImpactScore]: An iterator over ImpactScore objects for the company.
        """
        return StaticListIterator(
            self.client,
            f"/companies/{company_id}/climate/impacts",
            {
                "pathway": pathway,
                "horizon": horizon,
            },
            item_class=ImpactScore,
        )

    async def get_company_impact_scores_async(
        self, company_id: str, pathway: Pathway, horizon: HorizonYear
    ) -> StaticListIterator[ImpactScore]:
        """
        Get the impact scores for a company asynchronously.

        Parameters:
            company_id (str): The unique identifier of the company.
            pathway (Pathway): Climate scenario pathway powered by Climate Earth Digital Twin.
            horizon (HorizonYear): Climatology year representing a decadal period.

        Returns:
            StaticListIterator[ImpactScore]: An asynchronous iterator over ImpactScore objects for the company.
        """
        return StaticListIterator(
            self.client,
            f"/companies/{company_id}/climate/impacts",
            {
                "pathway": pathway,
                "horizon": horizon,
            },
            item_class=ImpactScore,
        )

    def list_company_asset_climate_scores(
        self,
        company_id: str,
        pathway: Pathway,
        horizon: HorizonYear,
        **extra_params: Any,
    ) -> PaginatedIterator[AssetClimateScore]:
        """
        Get the climate scores for all assets of a company.

        Parameters:
            company_id (str): The unique identifier of the company.
            pathway (Pathway): Climate scenario pathway powered by Climate Earth Digital Twin.
            horizon (HorizonYear): Climatology year representing a decadal period.
            **extra_params (Any): Additional parameters to pass to the API.

        Returns:
            PaginatedIterator[AssetClimateScore]: An iterator over AssetClimateScore objects for the company's assets.
        """
        params: Dict[str, Any] = {}
        params["pathway"] = pathway
        params["horizon"] = horizon
        params["metric"] = (
            "dcr_score,cvar_99,var_99,cvar_95,var_95,cvar_50,var_50,expected_impact"
        )
        params.update(extra_params)
        return PaginatedIterator(
            self.client,
            f"/companies/{company_id}/assets/climate/scores",
            params,
            item_class=AssetClimateScore,
        )

    async def list_company_asset_climate_scores_async(
        self,
        company_id: str,
        pathway: Pathway,
        horizon: HorizonYear,
        **extra_params: Any,
    ) -> AsyncPaginatedIterator[AssetClimateScore]:
        """
        Get the climate scores for all assets of a company asynchronously.

        Parameters:
            company_id (str): The unique identifier of the company.
            pathway (Pathway): Climate scenario pathway powered by Climate Earth Digital Twin.
            horizon (HorizonYear): Climatology year representing a decadal period.
            **extra_params (Any): Additional parameters to pass to the API.

        Returns:
            AsyncPaginatedIterator[AssetClimateScore]: An asynchronous iterator over AssetClimateScore objects for the company's assets.
        """
        params: Dict[str, Any] = {}
        params["pathway"] = pathway
        params["horizon"] = horizon
        params["metric"] = (
            "dcr_score,cvar_99,var_99,cvar_95,var_95,cvar_50,var_50,expected_impact"
        )
        params.update(extra_params)
        return AsyncPaginatedIterator(
            self.client,
            f"/companies/{company_id}/assets/climate/scores",
            params,
            item_class=AssetClimateScore,
        )

    def list_company_asset_impact_scores(
        self,
        company_id: str,
        pathway: Pathway,
        horizon: HorizonYear,
        **extra_params: Any,
    ) -> PaginatedIterator[AssetImpactScore]:
        """
        Get the impact scores for all assets of a company.

        Parameters:
            company_id (str): The unique identifier of the company.
            pathway (Pathway): Climate scenario pathway powered by Climate Earth Digital Twin.
            horizon (HorizonYear): Climatology year representing a decadal period.
            **extra_params (Any): Additional parameters to pass to the API.

        Returns:
            PaginatedIterator[AssetImpactScore]: An iterator over AssetImpactScore objects for the company's assets.
        """
        params: Dict[str, Any] = {}
        params["pathway"] = pathway
        params["horizon"] = horizon
        params["metric"] = (
            "dcr_score,cvar_99,var_99,cvar_95,var_95,cvar_50,var_50,expected_impact"
        )
        params.update(extra_params)
        return PaginatedIterator(
            self.client,
            f"/companies/{company_id}/assets/climate/impacts",
            params,
            item_class=AssetImpactScore,
            df_transform=lambda asset: [
                {
                    "asset_id": asset.asset_id,
                    **{
                        f"index_{risk.index_name}": risk.index_impact_cvar_50
                        for risk in asset.index_risks
                    },
                }
            ],
        )

    async def list_company_asset_impact_scores_async(
        self,
        company_id: str,
        pathway: Pathway,
        horizon: HorizonYear,
        **extra_params: Any,
    ) -> AsyncPaginatedIterator[AssetImpactScore]:
        """
        Get the impact scores for all assets of a company asynchronously.

        Parameters:
            company_id (str): The unique identifier of the company.
            pathway (Pathway): Climate scenario pathway powered by Climate Earth Digital Twin.
            horizon (HorizonYear): Climatology year representing a decadal period.
            **extra_params (Any): Additional parameters to pass to the API.

        Returns:
            AsyncPaginatedIterator[AssetImpactScore]: An asynchronous iterator over AssetImpactScore objects for the company's assets.
        """
        params: Dict[str, Any] = {}
        params["pathway"] = pathway
        params["horizon"] = horizon
        params["metric"] = (
            "dcr_score,cvar_99,var_99,cvar_95,var_95,cvar_50,var_50,expected_impact"
        )
        params.update(extra_params)
        return AsyncPaginatedIterator(
            self.client,
            f"/companies/{company_id}/assets/climate/impacts",
            params,
            item_class=AssetImpactScore,
            df_transform=lambda asset: [
                {
                    "asset_id": asset.asset_id,
                    **{
                        f"index_{risk.index_name}": risk.index_impact_cvar_50
                        for risk in asset.index_risks
                    },
                }
            ],
        )

    def aggregate_company_asset_climate_scores_by_country(
        self, company_id: str, pathway: Pathway, horizon: HorizonYear
    ) -> StaticListIterator[CountryClimateScore]:
        """
        Get the climate scores for all assets of a company aggregated by country.

        Parameters:
            company_id (str): The unique identifier of the company.
            pathway (Pathway): Climate scenario pathway powered by Climate Earth Digital Twin.
            horizon (HorizonYear): Climatology year representing a decadal period.

        Returns:
            StaticListIterator[CountryClimateScore]: An iterator over CountryClimateScore objects, aggregated by country.
        """
        return StaticListIterator(
            self.client,
            f"/companies/{company_id}/assets/climate/scores/aggregation",
            {
                "by": "country",
                "pathway": pathway,
                "horizon": horizon,
                "metric": "dcr_score,cvar_99,var_99,cvar_95,var_95,cvar_50,var_50,expected_impact",
            },
            item_class=CountryClimateScore,
        )

    async def aggregate_company_asset_climate_scores_by_country_async(
        self, company_id: str, pathway: Pathway, horizon: HorizonYear
    ) -> StaticListIterator[CountryClimateScore]:
        """
        Get the climate scores for all assets of a company aggregated by country asynchronously.

        Parameters:
            company_id (str): The unique identifier of the company.
            pathway (Pathway): Climate scenario pathway powered by Climate Earth Digital Twin.
            horizon (HorizonYear): Climatology year representing a decadal period.

        Returns:
            StaticListIterator[CountryClimateScore]: An asynchronous iterator over CountryClimateScore objects, aggregated by country.
        """
        return StaticListIterator(
            self.client,
            f"/companies/{company_id}/assets/climate/scores/aggregation",
            {
                "by": "country",
                "pathway": pathway,
                "horizon": horizon,
                "metric": "dcr_score,cvar_99,var_99,cvar_95,var_95,cvar_50,var_50,expected_impact",
            },
            item_class=CountryClimateScore,
        )

    def aggregate_company_asset_impact_scores_by_country(
        self, company_id: str, pathway: Pathway, horizon: HorizonYear
    ) -> StaticListIterator[CountryImpactScore]:
        """
        Get the impact scores for all assets of a company aggregated by country.

        Parameters:
            company_id (str): The unique identifier of the company.
            pathway (Pathway): Climate scenario pathway powered by Climate Earth Digital Twin.
            horizon (HorizonYear): Climatology year representing a decadal period.

        Returns:
            StaticListIterator[CountryImpactScore]: An iterator over CountryImpactScore objects, aggregated by country.
        """
        return StaticListIterator(
            self.client,
            f"/companies/{company_id}/assets/climate/impacts/aggregation",
            {
                "by": "country",
                "pathway": pathway,
                "horizon": horizon,
                "metric": "dcr_score,cvar_99,var_99,cvar_95,var_95,cvar_50,var_50,expected_impact",
            },
            item_class=CountryImpactScore,
        )

    async def aggregate_company_asset_impact_scores_by_country_async(
        self, company_id: str, pathway: Pathway, horizon: HorizonYear
    ) -> StaticListIterator[CountryImpactScore]:
        """
        Get the impact scores for all assets of a company aggregated by country asynchronously.

        Parameters:
            company_id (str): The unique identifier of the company.
            pathway (Pathway): Climate scenario pathway powered by Climate Earth Digital Twin.
            horizon (HorizonYear): Climatology year representing a decadal period.

        Returns:
            StaticListIterator[CountryImpactScore]: An asynchronous iterator over CountryImpactScore objects, aggregated by country.
        """
        return StaticListIterator(
            self.client,
            f"/companies/{company_id}/assets/climate/impacts/aggregation",
            {
                "by": "country",
                "pathway": pathway,
                "horizon": horizon,
                "metric": "dcr_score,cvar_99,var_99,cvar_95,var_95,cvar_50,var_50,expected_impact",
            },
            item_class=CountryImpactScore,
        )

    def aggregate_company_asset_climate_scores_by_asset_type(
        self, company_id: str, pathway: Pathway, horizon: HorizonYear
    ) -> StaticListIterator[AssetTypeClimateScore]:
        """
        Get the climate scores for all assets of a company aggregated by asset type.

        Parameters:
            company_id (str): The unique identifier of the company.
            pathway (Pathway): Climate scenario pathway powered by Climate Earth Digital Twin.
            horizon (HorizonYear): Climatology year representing a decadal period.

        Returns:
            StaticListIterator[AssetTypeClimateScore]: An iterator over AssetTypeClimateScore objects, aggregated by asset type.
        """
        return StaticListIterator(
            self.client,
            f"/companies/{company_id}/assets/climate/scores/aggregation",
            {
                "by": "asset_type",
                "pathway": pathway,
                "horizon": horizon,
                "metric": "dcr_score,cvar_99,var_99,cvar_95,var_95,cvar_50,var_50,expected_impact",
            },
            item_class=AssetTypeClimateScore,
        )

    async def aggregate_company_asset_climate_scores_by_asset_type_async(
        self, company_id: str, pathway: Pathway, horizon: HorizonYear
    ) -> StaticListIterator[AssetTypeClimateScore]:
        """
        Get the climate scores for all assets of a company aggregated by asset type asynchronously.

        Parameters:
            company_id (str): The unique identifier of the company.
            pathway (Pathway): Climate scenario pathway powered by Climate Earth Digital Twin.
            horizon (HorizonYear): Climatology year representing a decadal period.

        Returns:
            StaticListIterator[AssetTypeClimateScore]: An asynchronous iterator over AssetTypeClimateScore objects, aggregated by asset type.
        """
        return StaticListIterator(
            self.client,
            f"/companies/{company_id}/assets/climate/scores/aggregation",
            {
                "by": "asset_type",
                "pathway": pathway,
                "horizon": horizon,
                "metric": "dcr_score,cvar_99,var_99,cvar_95,var_95,cvar_50,var_50,expected_impact",
            },
            item_class=AssetTypeClimateScore,
        )

    def aggregate_company_asset_impact_scores_by_asset_type(
        self, company_id: str, pathway: Pathway, horizon: HorizonYear
    ) -> StaticListIterator[AssetTypeImpactScore]:
        """
        Get the impact scores for all assets of a company aggregated by asset type.

        Parameters:
            company_id (str): The unique identifier of the company.
            pathway (Pathway): Climate scenario pathway powered by Climate Earth Digital Twin.
            horizon (HorizonYear): Climatology year representing a decadal period.

        Returns:
            StaticListIterator[AssetTypeImpactScore]: An iterator over AssetTypeImpactScore objects, aggregated by asset type.
        """
        return StaticListIterator(
            self.client,
            f"/companies/{company_id}/assets/climate/impacts/aggregation",
            {
                "by": "asset_type",
                "pathway": pathway,
                "horizon": horizon,
                "metric": "dcr_score,cvar_99,var_99,cvar_95,var_95,cvar_50,var_50,expected_impact",
            },
            item_class=AssetTypeImpactScore,
        )

    async def aggregate_company_asset_impact_scores_by_asset_type_async(
        self, company_id: str, pathway: Pathway, horizon: HorizonYear
    ) -> StaticListIterator[AssetTypeImpactScore]:
        """
        Get the impact scores for all assets of a company aggregated by asset type asynchronously.

        Parameters:
            company_id (str): The unique identifier of the company.
            pathway (Pathway): Climate scenario pathway powered by Climate Earth Digital Twin.
            horizon (HorizonYear): Climatology year representing a decadal period.

        Returns:
            StaticListIterator[AssetTypeImpactScore]: An asynchronous iterator over AssetTypeImpactScore objects, aggregated by asset type.
        """
        return StaticListIterator(
            self.client,
            f"/companies/{company_id}/assets/climate/impacts/aggregation",
            {
                "by": "asset_type",
                "pathway": pathway,
                "horizon": horizon,
                "metric": "dcr_score,cvar_99,var_99,cvar_95,var_95,cvar_50,var_50,expected_impact",
            },
            item_class=AssetTypeImpactScore,
        )



from .companies import Companies
from .assets import Assets
from .base import BaseClient
from .markets import Markets
from .climate import Climate


class APIClient(BaseClient):
    def __init__(
        self,
        api_key: str | None = None,
        timeout: float = 15.0,
        base_url: str | None = None,
    ):
        super().__init__(api_key, timeout, base_url)

        # Company endpoints
        self.companies = Companies(self)

        # Asset endpoints
        self.assets = Assets(self)

        # Market endpoints
        self.markets = Markets(self)

        # Climate endpoints
        self.climate = Climate(self)



from pydantic import BaseModel
from typing import Literal

Pathway = Literal[
    "SV",
    "ssp126",
    "ssp245",
    "ssp370",
    "ssp585",
    "<2 degrees",
    "2-3 degrees",
    "3-4 degrees",
    ">4 degrees",
]

HorizonYear = Literal[
    2025,
    2030,
    2035,
    2040,
    2045,
    2050,
    2055,
    2060,
    2065,
    2070,
    2075,
    2080,
    2085,
    2090,
    2095,
]


class Company(BaseModel):
    """
    A company is an entity that has assets and identifying information.
    """

    id: str
    name: str
    slug: str
    headquarters_address: str
    organization_id: str | None = None
    market_cap: int
    annual_revenue: int
    headquarters_country: str
    sector: str
    isin_codes: list[str]
    figi_codes: list[str]
    cik_code: str
    lei_code: str
    stock_tickers: list[str]
    is_grandparent: bool | None = None
    data_generated_at: str | None = None
    data_generation_status: str | None = None
    created_at: str
    updated_at: str


class Asset(BaseModel):
    """
    An asset represents aphysical asset that is subject to climate risk.
    """

    id: str
    name: str
    asset_type: str
    city: str
    state: str
    latitude: float
    longitude: float
    building_footprint: float
    asset_value: int
    address: str
    hex_id: int
    country: str
    ipcc_region: str
    materiality_score: float
    created_at: str
    updated_at: str


class MarketIndex(BaseModel):
    """
    A market index is a collection of companies.
    """

    id: str
    name: str
    sectors: list[str]
    created_at: str
    updated_at: str
    organization_id: str | None = None


class ClimateScore(BaseModel):
    """
    The cimate risk metrics that represent the likelihood of a company or asset to be impacted by climate risk.
    """

    dcr_score: float | None = None
    expected_impact: float | None = None
    cvar_99: float | None = None
    cvar_95: float | None = None
    cvar_50: float | None = None
    var_99: float | None = None
    var_95: float | None = None
    var_50: float | None = None


class ImpactScore(BaseModel):
    """
    The impact risk metrics that represent the potential impact of a company or asset to be impacted by climate risk.
    These metrics represent an individual risk factor and its attribution to the total climate risk metrics.
    """

    index_name: str
    index_impact_cvar_50: float | None = None
    index_impact_cvar_95: float | None = None
    index_impact_cvar_99: float | None = None
    index_impact_var_50: float | None = None
    index_impact_var_95: float | None = None
    index_impact_var_99: float | None = None
    index_impact_expected: float | None = None
    index_attribution_expected: float | None = None
    index_attribution_var_99: float | None = None
    index_attribution_var_95: float | None = None
    index_attribution_var_50: float | None = None
    index_attribution_cvar_99: float | None = None
    index_attribution_cvar_95: float | None = None
    index_attribution_cvar_50: float | None = None


class CountryClimateScore(ClimateScore):
    """
    Climate risk metrics aggregated for a country.
    """

    asset_count: int
    country: str


class CountryImpactScore(ImpactScore):
    """
    Impact risk metrics aggregated for a country.
    """

    asset_count: int
    country: str


class AssetTypeClimateScore(ClimateScore):
    """
    Climate risk metrics aggregated for an asset type.
    """

    asset_count: int
    asset_type: str


class AssetTypeImpactScore(ImpactScore):
    """
    Impact risk metrics aggregated for an asset type.
    """

    asset_count: int
    asset_type: str


class AssetClimateScore(ClimateScore):
    """
    Climate risk metrics for an asset.
    """

    asset_id: str
    asset_type: str
    country: str
    state: str
    city: str
    address: str


class AssetImpactScore(BaseModel):
    """
    Impact risk metrics for an asset.
    """

    asset_id: str
    index_risks: list[ImpactScore]



from typing import Any, Literal

from .base import BaseClient
from .types import Asset, Company
from .pagination import PaginatedIterator, AsyncPaginatedIterator


class Assets:
    def __init__(self, client: BaseClient):
        self.client = client

    def get_asset(self, asset_id: str) -> Asset:
        """
        Get an asset by its unique ID.

        Parameters:
            asset_id (str): The unique identifier of the asset.

        Returns:
            Asset: The Asset object.
        """
        response = self.client._request_sync("GET", f"/assets/{asset_id}")
        return Asset(**response)

    async def get_asset_async(self, asset_id: str) -> Asset:
        """
        Get an asset by its unique ID asynchronously.

        Parameters:
            asset_id (str): The unique identifier of the asset.

        Returns:
            Asset: The Asset object.
        """
        response = await self.client._request_async("GET", f"/assets/{asset_id}")
        return Asset(**response)

    def list_assets(
        self,
        **extra_params: Any,
    ) -> PaginatedIterator[Asset]:
        """
        List all assets.

        Parameters:
            **extra_params (Any): Additional parameters to pass to the API.

        Returns:
            PaginatedIterator[Asset]: An iterator over Asset objects.
        """
        return PaginatedIterator(self.client, "/assets", extra_params, item_class=Asset)

    async def list_assets_async(
        self,
        **extra_params: Any,
    ) -> AsyncPaginatedIterator[Asset]:
        """
        List all assets asynchronously.

        Parameters:
            **extra_params (Any): Additional parameters to pass to the API.

        Returns:
            AsyncPaginatedIterator[Asset]: An asynchronous iterator over Asset objects.
        """
        return AsyncPaginatedIterator(
            self.client, "/assets", extra_params, item_class=Asset
        )

    def get_asset_owner(self, asset_id: str) -> Company:
        """
        Get the company that owns an asset.

        Parameters:
            asset_id (str): The unique identifier of the asset.

        Returns:
            Company: The Company object that owns the asset.
        """
        response = self.client._request_sync("GET", f"/assets/{asset_id}/ownership")
        return Company(**response)

    async def get_asset_owner_async(self, asset_id: str) -> Company:
        """
        Get the company that owns an asset asynchronously.

        Parameters:
            asset_id (str): The unique identifier of the asset.

        Returns:
            Company: The Company object that owns the asset.
        """
        response = await self.client._request_async(
            "GET", f"/assets/{asset_id}/ownership"
        )
        return Company(**response)

    def search_assets(
        self,
        query: str,
        scope: Literal["public", "company", "organization"] = "public",
        company_id: str | None = None,
        **extra_params: Any,
    ) -> PaginatedIterator[Asset]:
        """
        Search for assets.

        Parameters:
            query (str): The search query string. Asset names and addresses are searched for.
            scope (Literal["public", "company", "organization"]): The scope of the search.
                   "public" is the default scope and searches all available assets in VELO.
                   "organization" searches all private assets uploaded to the organization.
                   If "company" is selected, `company_id` must also be provided.
            company_id (Optional[str]): The ID of the company to scope the search to.
                        Required if `scope` is "company".
            **extra_params (Any): Additional parameters to pass to the API.

        Returns:
            PaginatedIterator[Asset]: A paginated iterator of assets matching the search criteria.
        """
        params = {
            "query": query,
            "scope": scope,
            "company_id": company_id,
            **extra_params,
        }
        params = {k: v for k, v in params.items() if v is not None}
        return PaginatedIterator(
            self.client, "/assets/search", params, item_class=Asset
        )

    async def search_assets_async(
        self,
        query: str,
        scope: Literal["public", "company", "organization"] = "public",
        company_id: str | None = None,
        **extra_params: Any,
    ) -> AsyncPaginatedIterator[Asset]:
        """
        Search for assets asynchronously.

        Parameters:
            query (str): The search query string. Asset names and addresses are searched for.
            scope (Literal["public", "company", "organization"]): The scope of the search.
                   "public" is the default scope and searches all available assets in VELO.
                   "organization" searches all private assets uploaded to the organization.
                   If "company" is selected, `company_id` must also be provided.
            company_id (Optional[str]): The ID of the company to scope the search to.
                        Required if `scope` is "company".
            **extra_params (Any): Additional parameters to pass to the API.

        Returns:
            AsyncPaginatedIterator[Asset]: An asynchronous paginated iterator of assets matching the search criteria.
        """
        params = {
            "query": query,
            "scope": scope,
            "company_id": company_id,
            **extra_params,
        }
        params = {k: v for k, v in params.items() if v is not None}
        return AsyncPaginatedIterator(
            self.client, "/assets/search", params, item_class=Asset
        )
